#!/usr/bin/perl
#
# Minimalist - MINImalist MAiling LIST manager.
# Copyright (c) 1999 Vladimir Litovka <doka@webest.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 675
# Mass Ave, Cambridge, MA 02139, USA.

$version = "1.3.3";
$config = "/usr/local/etc/minimalist.conf";

# Lists' status bits
$OPEN = 0;
$RO = 1;
$CLOSED = 2;
$MANDATORY = 4;

$sendmail = '/usr/sbin/sendmail';
$domain = `hostname --fqdn`; chomp $domain;
$directory = '/var/spool/minimalist';
$admin = "root\@$domain";
$listpwd = $userpwd = '';
$security = 'careful';
$archive = 'no';
$status = $OPEN;
$copy_sender = 'yes';
$reply_to_list = 'no';
$modify_subject = 'yes';
$maxusers = 0;

#####################################################
# Various regular expressions
#
# for matching rounding spaces

$spaces = '^[\s\t]*([\w\.\,\@\/]*)[\s\t]*$';

# for parsing two forms of mailing addresses:
#
# 1st form: Vladimir Litovka <doka@webest.com>
# 2nd form: doka@webest.com (Vladimir Litovka)

$first = '.*<(.*\@.*\..*)>.*';
$second = '(.*\@.*\..*)\(.*\).*';

# Help
$helpmsg = 
	"This is Minimalist server. All commands must appear in the SUBJECT. ".
	"Supported commands are:\n\n".
	"subscribe <list> [<email>] :\n\tsubscribe user to <list>\n\n".
	"unsubscribe <list> [<email>] :\n\tunsubscribe user from <list>\n\n".
	"info [<list>] :\n\tgives you information about existing lists or about <list>\n\n".
	"list <list> :\n\tgives you listing of subscribed to <list> users\n\n".
	"help :\n\tThis message\n\n".
	"Note, that <email> can be used only by trusted users. Otherwise, address ".
	"in field From: will be processed. Also, command LIST can be used only by trusted ".
	"users. Trusted users must supply password in first line of message's body, as following:\n\n".
	"*password: list_password\n\n".
	"followed by any number of empty rows. Password will be removed from message, of cource.";

########################################################
# >>>>>>>>>>>>>>> SELF - CONFIGURING <<<<<<<<<<<<<<<<< #
########################################################

read_config($config, "global");

# Admin always in trusted users
$trusted[@trusted] = $admin;

$me = "minimalist\@$domain";

#######################################################
# >>>>>>>>>>>>>>>> CHECK CONFIGURATION <<<<<<<<<<<<<< #
#######################################################

if ($ARGV[0] eq '-') {
  print "\nMinimalist v$version, pleasured to meet you.\n\n".
	"================= Global configuration ================\n".
	"Directory: $directory\n".
	"Admin: $admin\n";
  die "$directory doesn't exist!\n" if (! -d $directory);

  while ( $ARGV[0] ) {
    $list = $ARGV[0];
    if ($list ne '-') {
      if (!chdir("$directory/$list")) {
        print "\nThere isn't such list \U$list\E\n";
        shift; next;
       }
      read_config("config");
      print "================= \U$list\E ================\n";
     }
    print "Sendmail: $sendmail\n".
	"Domain: $domain\n".
	"Password: ".($listpwd ? "yes, list protected" : "none")."\n".
	"Security: $security\n".
	"Archiving: $archive\n".
	"Status: $status\n".
	"Copy to sender: $copy_sender\n".
	"Reply-To list: $reply_to_list\n".
	"Modify subject: $modify_subject\n".
	"Maximal users per list: ".($maxusers ? $maxusers : "unlimited")."\n".
	"Trusted users:\n";
    foreach (@trusted) { print "\t + ".$_."\n"; }
    print "\n";
    
    # Various checks
    die "$sendmail doesn't exist!\n" if (! -x $sendmail);
    die "Illegal domain '$domain'\n" if ($domain !~ /^([a-zA-Z0-9\-]+\.)+[a-zA-Z0-9\-]+$/);
    die "Illegal security level '$security'\n" if ($security !~ /^none$|^careful$/i);
    die "Illegal archiving strategy '$archive'\n" if ($archive !~ /^no$|^daily$|^monthly$|^yearly$/i);
    die "Illegal 'copy to sender' value '$copy_sender'\n" if ($copy_sender !~ /^yes$|^no$/i);
    die "Illegal 'modify subject' value '$modify_subject'\n" if ($modify_subject !~ /^yes$|^no$|^more$/i);
    die "Illegal 'reply-to list' value '$reply_to_list'\n" if ($reply_to_list !~ /^yes$|^no$|\@/i);

    shift;
   }

  exit 0;
 }

########################################################
# >>>>>>>>>>>>>>>>>> START HERE <<<<<<<<<<<<<<<<<<<<<< #
########################################################

$list = $ARGV[0];

while (<STDIN>) { $message .= $_ }

($header, $body) = split(/\n\n/, $message, 2); $header .= "\n";
@headers = split (/\n/, $header);

# Get password if it's there and strip this line from message's body
if ($body =~ s/(^[\s\t]*\*password: )(.*)\n+(.*)/$3/i) {
  $userpwd = $2;
 }

#####################################################
# Mail FROM, but REPLY-TO

$from = ''; $reply = ''; $subject = '';

foreach (@headers){
  if    ( /^from:[ ]*/i && ($from eq '') ) { $from = $_; }
  elsif ( /^reply-to:[ ]*/i && ($reply eq '') ) { $reply = $_; }
  elsif ( /^x-sender:[ ]*/i ) { $xsender = $_; }
 }
$from =~ s/^from:[ ]*//ig;
$reply =~ s/^reply-to:[ ]*//ig;
$xsender =~ s/^x-sender:[ ]*//ig;

exit 0 if (($xsender eq $me) || ($from eq $me));	# LOOP detected
exit 0 if ($from =~ /MAILER-DAEMON/i);			# ignore messages from MAILER-DAEMON

($reply eq '') ? ($mailto = $from) : ($mailto = $reply);

# Preparing From:
if ($from =~ /<.*>/) { $from =~ s/$first/$1/og; }
elsif ($from =~ /\(.*\)/) { $from =~ s/$second/$1/og; }
$from =~ s/$spaces/$1/g;
$from = lc($from);

# Get (multiline) subject
$header =~ /subject: (.*\n([\s\t]+.*\n)*)/i;
$subject = $1; chomp $subject;

#####################################################
# Message to list
#
if ($list ne '') {

 chdir("$directory/$list");
 read_config("config");

 open LIST, "list";
 while (<LIST>) { $users .= $_; }
 close LIST;
 @members = split ("\n", $users);
 
 $subject =~ s/^[ \t]*//g;
 if ($modify_subject ne 'no') {
   if ($modify_subject eq 'more') { $subject =~ s/^Re:[ \t]*//ig }
   if ($subject !~ /^(Re: )?\[$list\] /i) {
     $subject = "[$list] ".$subject;
    }
  }
  
 if (($security eq 'careful') && !grep(/^$from$/i, @members)) {
   $msg = "From: $me\n".
          "To: $mailto\n"."Subject: $subject\n\n".
          "ERROR:\n\tYou ($from) are not subscribed to this list.\n\n".
	  "SOLUTION:\n\tSend a message to minimalist\@$domain with a subject\n".
	  "\tof 'info' (no quotes) for information about subscribing.\n\n".
          "===========================================================================\n\n".$body;
  } 
 elsif (($status & $RO) && (!grep(/^$from$/i, @trusted) || ($userpwd ne $listpwd))) {
   $msg = "From: $me\n".
          "To: $mailto\n"."Subject: $subject\n\n".
	  "ERROR:\n\tYou are not allowed to write to this list.\n\n".
	  "===========================================================================\n\n".$body;
  }
 else {

   # OK, all checks done. Now send message to members except those who already
   # received this message directly from sender

   # Get all recipients from To: and Cc: (expect multi-line)
   $header =~ /((.*\n)*)to: (.*\n([\s\t]+.*\n)*)/i; $rc = $3;
     @recip = split(/[\n\s\t,]+/, $rc);
   $header =~ /((.*\n)*)cc: (.*\n([\s\t]+.*\n)*)/i; $rc = $3;
     push (@recip, split(/[\n\s\t,]+/, $rc));

   foreach (@recip) {
     if ( /<.*>/ ) { s/$first/$1/g; }
     elsif ( /\(.*\)/ ) { s/$second/$1/g; }
     s/$spaces/$1/g;
     push (@rcpts, lc($_));
    }

   if ($copy_sender eq 'no') { push (@rcpts, $from) }

   @members = sort @members; @rcpts = sort @rcpts;

   for ($r=0, $m=0; $m < @members; ) {
     if ($members[$m] lt $rcpts[$r] || $r >= @rcpts) {
       push (@recipients, $members[$m]);
       $m++; }
     elsif ($members[$m] eq $rcpts[$r]) { $r++; $m++; }
     elsif ($members[$m] gt $rcpts[$r]) { $r++ };
    }

   archive() if ($archive ne 'no');		# Archive message if desired

   # Remove conflicting X- headers
   $header =~ s/x-mailing-list.*\n([\s\t]+.*\n)*//ig;
   $header =~ s/x-sender: .*\n([\s\t]+.*\n)*//ig;

   # Remove all recipients - they already received message
   $header =~ s/((.*\n)*?)(to:|cc:) (.*\n([\s\t]+.*\n)*)/$1/ig;

   chomp $header;
   $header .= "\nTo: $list\@$domain\n";		# Mail to himself, loop is handling

   if ($reply_to_list ne 'no') {
     $header =~ s/((.*\n)*?)reply-to: .*\n([\s\t]+.*\n)*/$1/ig;
     if ($reply_to_list eq 'yes') { $header .= "Reply-To: $list\@$domain\n"; }
     else { $header .= "Reply-To: $reply_to_list\n"; }
    }

   if ($modify_subject ne 'no') {
     $header =~ s/((.*\n)*?)subject: .*\n([\s\t]+.*\n)*/$1/ig;
     $header .= "Subject: $subject\n";
    }

   foreach (@recipients) {
     if ($z == 2) { $bcc .= "\n  "; $z = 0; }
     if ($_ ne '') { $bcc .= "$_,"; $z++; }
    }
   if ($bcc ne '') {
     chop $bcc;
     $header .= "Bcc: $bcc\n";			# Bcc to subscribers
    }

   $header .= "X-Sender: $me\n".
	      "X-Mailing-List-Server: Minimalist v$version\n".
   	      "X-Mailing-List: $list\@$domain\n".
	      "X-Mailing-List-Owner: $list-owner\@$domain\n\n";

   $msg = $header.$body.read_info("footer");

   $envelope_sender = "-f $me";
  }
} else {

#####################################################
# Message to minimalist
#
# Allowed commands:
#	subscribe <list> [<e-mail>]
#	unsubscribe <list> [<e-mail>]
#	info [<list>]
#	list <list>
#	help

 $list = ''; $email = '';
 ($cmd, $list, $email) = split (/ /, $subject, 3);
 $cmd =~ s/$spaces/$1/g; $cmd = lc($cmd);
 $list =~ s/$spaces/$1/g; $list = lc($list);

 if ($email ne '') {
   if ($email =~ /<.*>/) { $email =~ s/$first/$1/og; }
   elsif ($email =~ /\(.*\)/) { $email =~ s/$second/$1/og; }
   $email =~ s/$spaces/$1/g;
   $email = lc($email);
  }

 $msg = "From: $me\n".
	"To: $mailto\n".
	"Subject: Re: $subject\n";
 
 if ($cmd eq 'help') {
   $msg .= "\n".$helpmsg;
  }

 elsif ( ($list ne '') && (! -d "$directory/$list") ) {
   $msg .= "\nERROR:\n\tThere is no such list \U$list\E here.\n\nSOLUTION:\n".
	   "\tSend a message to minimalist\@$domain with a subject\n".
	   "\tof 'info' (no quotes) for a list of available mailing lists.";
  }

 elsif ( (($cmd eq 'subscribe') || ($cmd eq 'unsubscribe')) && ($list ne '') ) {
   $melist = "$list\@$domain";
   if (($from ne $me) && ($from ne $melist) && ($email ne $me) && ($email ne $melist)) {
     chdir "$directory/$list";
     read_config("config");
     $owner = "$list-owner\@$domain";

     if (grep(/^$from$/i, @trusted) && ($userpwd eq $listpwd)) { eval("$cmd(1)"); }
     elsif (($email ne '') && ($email ne $from)) {
       $msg .= "\nERROR:\n\tYou aren't allowed to subscribe other persons.\n\nSOLUTION:\n\tNone.";
      }
     elsif (($cmd eq 'subscribe') && ($status & $CLOSED)) {
       $msg .= "\nERROR:\n\tSorry, this list is closed for you.\n\n".
	       "SOLUTION:\n\tAre you unsure? Please, complain to $owner";
      }
     elsif (($cmd eq 'unsubscribe') && ($status & $MANDATORY)) {
       $msg .= "\nERROR:\n\tSorry, this list is mandatory for you.\n\n".
	       "SOLUTION:\n\tAre you unsure? Please, complain to $owner";
      }
     else {
       eval("$cmd(0)");
      }
    }	# loop preventing
  }	# subscribe/unsubscribe

 elsif ($cmd eq 'info') {
   if ($list ne '') {
     $msg .= "\nThere is available information about \U$list\E\n\n";
     $msg .= read_info("$directory/$list/info");
    }
   else {
     $msg .= "\nThere are mailing lists available at $domain:\n\n";
     if (open(INFO, "$directory/lists.lst")) {
       while (<INFO>) {
         if (! /^#/) { $msg .= $_ }
	}
       close INFO;
      }
    }
  }
 elsif (($cmd eq 'list') && ($list ne '')) {
   chdir "$directory/$list";
   read_config("config");

   if (grep(/^$from$/i, @trusted) && ($userpwd eq $listpwd)) {
     $msg .= "\nUsers, subscribed to \U$list\E:\n\n";
     if (open(LIST, "list")) {
       while (<LIST>) { $msg .= $_ }
       close LIST;
      }
    }
   else {
     $msg .= "\nERROR:\n\tYou aren't allowed to get listing of subscribed users.\n\nSOLUTION:\n\tNone"; }
  }
 else {
   $msg .= "\nERROR:\n\tBad syntax or unknown instruction.\n\nSOLUTION:\n\n".$helpmsg; }
}

open MAIL, "| $sendmail -t $envelope_sender";
print MAIL $msg."\n\n";
close MAIL;

exit 0;

#####################################################################################
# Supplementary functions
#

#################### SUBSCRIBE #####################
sub subscribe {

 my ($trusteduser) = @_;
 my ($cc);

 if ($email ne '') {
   $cc = "$email,";
   $from = $email;
  }

 if (open LIST, "list") {
   while (<LIST>) { $users .= $_; }
   close LIST;
   @members = split ("\n", $users);
   if (grep(/^$from$/i, @members)) {
     $deny = 1;
     $cause = "you already subscribed to \U$list\E";
    }
   elsif ((! $trusteduser) && ($maxusers > 0) && (@members >= $maxusers)) {
     $deny = 1; $cc .= "$owner,";
     $cause = "there are already maximal count of subscribers ($maxusers) at \U$list\E";
    }
   else { open LIST, ">>list"; }
  }
 else { open LIST, ">list"; } 

 if ($cc) {
   chop $cc;
   $msg .= "Cc: $cc\n";
  }

 $msg .= "\nDear $from,\n\n";

 if (! $deny) {
   print LIST "$from\n";
   close LIST;
   $msg .= "\tYou have subscribed to \U$list\E successfully.\n".
           "\nPlease, note the following:\n\n".read_info("info");
  }
 else {
   $msg .= "You haven't subscribed to \U$list\E due to the following reason:\n\t* $cause.\n\n".
   	   "If you have comments or questions, please, send them to list's administrator $owner.\n";
  }
}

#################### UNSUBSCRIBE #####################
sub unsubscribe {
 if ($email ne '') {
   $msg .= "Cc: $email\n" ;
   $from = $email;
  }

 if (open LIST, "list") {
   while (<LIST>) { $users .= $_; }
   close LIST;

   $users =~ s/$from\n//g;
   open LIST, ">list";
   print LIST $users;
   close LIST;
  }

 $msg .= "\nUser $from has successfully unsubscribed.\n\nGood luck!";
}

######################### READ_CONFIG ############################
sub read_config {

my ($fname, $global) = @_;

if (open(CONF, $fname)) {
  while (<CONF>) {
    if($_ =~ /^#/) {
    } elsif (($_ =~ /^directory/i) && $global) {
     ($directive, $directory) = split(/=/, $_);
     chomp $directory; $directory =~ s/$spaces/$1/g;

    } elsif (($_ =~ /^admin/i) && $global) {
     ($directive, $admin) = split(/=/, $_);
     chomp $admin; $admin =~ s/$spaces/$1/g;

    } elsif ($_ =~ /^sendmail/i) {
     ($directive, $sendmail) = split(/=/, $_);
     chomp $sendmail; $sendmail =~ s/$spaces/$1/g;

    } elsif ($_ =~ /^domain/i) {
     ($directive, $domain) = split(/=/, $_);
     chomp $domain; $domain =~ s/$spaces/$1/g;

    } elsif ($_ =~ /^password/i) {
     ($directive, $listpwd) = split(/=/, $_);
     chomp $listpwd; $listpwd =~ s/$spaces/$1/g;

    } elsif ($_ =~ /^security/i) {
     ($directive, $security) = split(/=/, $_);
     chomp $security; $security =~ s/$spaces/$1/g;
     $security = lc($security);

    } elsif ($_ =~ /^archive/i) {
     ($directive, $archive) = split(/=/, $_);
     chomp $archive; $archive =~ s/$spaces/$1/g;
     $archive = lc($archive);

    } elsif ($_ =~ /^maxusers/i) {
     ($directive, $maxusers) = split(/=/, $_);
     chomp $maxusers; $maxusers =~ s/$spaces/$1/g;

    } elsif ($_ =~ /^status/i) {
     ($directive, $status) = split(/=/, $_);
     chomp $status; $status =~ s/$spaces/$1/g;
     $status = lc($status);

     # Calculate mask for status
     %strel = ("open", $OPEN, "ro", $RO, "closed", $CLOSED, "mandatory", $MANDATORY);
     @starr = split(/,/, $status);
     $status = 0;
     foreach (@starr) { $status += $strel{$_}; }

    } elsif ($_ =~ /^copy to sender/i) {
     ($directive, $copy_sender) = split(/=/, $_);
     chomp $copy_sender; $copy_sender =~ s/$spaces/$1/g;
     $copy_sender = lc($copy_sender);

    } elsif ($_ =~ /^reply-to list/i) {
     ($directive, $reply_to_list) = split(/=/, $_);
     chomp $reply_to_list; $reply_to_list =~ s/$spaces/$1/g;
     $reply_to_list = lc($reply_to_list);
     # In global config only 'yes' or 'no' allowed
     if ($global && ($reply_to_list ne 'yes')) { $reply_to_list = 'no'; }

    } elsif ($_ =~ /^modify subject/i) {
     ($directive, $modify_subject) = split(/=/, $_);
     chomp $modify_subject; $modify_subject =~ s/$spaces/$1/g;
     $modify_subject = lc($modify_subject);

    } elsif ($_ =~ /^trusted/i) {
     ($directive, $trust) = split(/=/, $_, 2);
     chomp $trust; $trust =~ s/[\s\t]*//g;
     @junk = split(':', $trust);
     foreach $s (@junk) {
       if ( $s =~ s/^\@// ) {	# Expand items, starting with '@'
	 if (open(IN, $s)) {
	   while (<IN>) {
	     chomp $_; $trusted[@trusted] = $_; }
	   close IN;
	  }
	}
       elsif ($s ne '') { $trusted[@trusted] = $s; }
      }
    }
  }
  close CONF;
 }
}

sub read_info {
 my ($fname) = @_;
 my ($tail);

 if (open(TAIL, $fname)) {
   while (<TAIL>) { $tail .= $_ }
   close TAIL;

   $tail =~ s/\\a/$admin/ig;
   $tail =~ s/\\d/$domain/ig;
   $tail =~ s/\\l/$list/ig;
   $tail =~ s/\\o/$list-owner\@$domain/ig;
  }

 $retval = $tail;
}

########################## ARCHIVE ###############################
sub archive {

 @date = localtime;
 $year = 1900 + $date[5];
 $month = 1 + $date[4];
 $day = $date[3];

 $path = "archive/";
 mkdir($path, 0755) if (! -d $path);

 @types = ("yearly", "monthly", "daily");
 %rel = ($types[0], $year, $types[1], $month, $types[2], $day);

 foreach $key (@types) {
   $path .= $rel{$key}."./";
   mkdir($path, 0755) if (! -d $path);
   last if ($key eq $archive);
  }

 if (open(NUM, $path."SEQUENCE")) {
   read NUM, $msgnum, 16;
   $msgnum = int($msgnum);
   close NUM;
  }
 else { $msgnum = 0 }

 open ARCHIVE, ">$path".$msgnum;
 print ARCHIVE $message."\n";
 close ARCHIVE;

 open NUM, ">$path"."SEQUENCE";
 print NUM $msgnum+1;
 close NUM;
}
